//------------------------------------------------------------------------------
//	SainSmart IIC/I2C/TWI 16*02 LCDt W[ For Arduino UNO MEGA R3
//	(HD44780/SC1602 LCDW[(16 x 2) + I2CExpander)@R}h
//	TPł́A2s\ɑ΂āAꂼ̍sɕ\܂B
//	{I16Œ\ŃXN[͂܂B
//
//	# lcd_exp [-a ] [-i] [-h] <text> [<line#>]
//
//	<text>	ASCII(SJIS:Ji)
//	<line#>	sԍF	0 or 3
//
//	option
//		-a <addr>:	I2C̃AhXiftHgF0x27)
//		-i:		LCD̏s
//		-h:		wv\
//
//	4s\ɑ΂āAꂼ̍sɕ\܂B
//	{I16Œ\ŃXN[͂܂B
//	J[\\܂B
//
//	RpC@
//		gcc -o lcd_exp lcd_exp.c -l bcm2835
//
//	copyright (c) 2016 Kazuhiro WATANABE All right reserved.
//	mailto:jj1req@ca.mbn.or.jp
//
//------------------------------------------------------------------------------

#include <bcm2835.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

// LN^tfBXv[ւ̐ڑ
//	P0	RS
//	P1	R_W
//	P2	E
//	P3	Backlight
//	P4	D4
//	P5	D5
//	P6	D6
//	P7	D7

#define LCD_LIGHT	0x08
#define LCD_E   	0x04
#define LCD_R_W		0x02
#define LCD_RS		0x01

#define LCD_WR		0x00
#define LCD_RD		0x01
#define LCD_COMMAND	0x00
#define LCD_DATA	0x01

//#define I2C_BASE_ADDR	0x3F
#define I2C_BASE_ADDR	0x27
#define CLK_DIV BCM2835_I2C_CLOCK_DIVIDER_2500

//
// Commands
// Function Set
//	0 0 1 DL N DH x x
//		DL : Interface data length control bit
//			1: 8-bit bus mode with MPU.
//			0: 4-bit bus mode with MPU.
//				When in 4-bit bus mode, it needs to transfer 4-bit data by two times.
//		N : Display line number control bit
//			1: 2-line display mode is set.
//			0: 1-line display mode.
//		DH : Double height font type control bit
//			1: double height mode(5x16 dot),RAM address can only use 00H~27H.
//			0: normal (5x8 dot).
#define LCD_FUNCTION_SET	0x28		// 4bit bus, 2 line, normal font
//
// Clear Display
//	0 0 0 0 0 0 0 1
#define LCD_CLEAR	0x01
//
// Display ON/OFF
//	0 0 0 0 1 D C B
//		D : Display ON/OFF control bit
//			1: entire display is turned on.
//			0: display is turned off, but display data is remained in DDRAM.
//		C : Cursor ON/OFF control bit
//			1: cursor is turned on.
//			0: cursor is disappeared in current display, but I/D register remains its data.
//		B : Cursor Blink ON/OFF control bit
//			1: cursor blink is on, that performs alternate between all the high data and display character at the cursor position.
//			0: blink is off.
#define LCD_ON		0x0c
#define LCD_OFF		0x08

// Command Recovery Time
#define LCD_RECOVERY	100LL	// 100us

#define ON		1
#define OFF		0


char	Backlight = 0x08;
uint8_t	I2C_addr;
//******************************************************************************
//*	I2CANZX̃G[v\܂B
//*
//* Function Name
//*	void get_reaosn( bcm2835I2CReasonCodes res, char *str )
//* Arguments
//*	bcm2835I2CReasonCodes res	error code
//*	char str			message buffer
//* Return
//*	none
//*
//******************************************************************************
void get_reaosn( bcm2835I2CReasonCodes res, char *str )
{
	switch( res )
	{
	case BCM2835_I2C_REASON_OK:
		strcpy( str, "Success" );
		break;
	case BCM2835_I2C_REASON_ERROR_NACK:
		strcpy( str, "Received a NACK" );
		break;
	case BCM2835_I2C_REASON_ERROR_CLKT:
		strcpy( str, "Received Clock Stretch Timeout" );
		break;
	case BCM2835_I2C_REASON_ERROR_DATA:
		strcpy( str, "Not all data is sent / received" );
		break;
	default:
		strcpy( str, "Unknown error" );
	}
}

//******************************************************************************
//*	LCDW[ɑ΂ăXg[u܂iE[qXg[uj
//*
//* Function Name
//*	void lcd_strobe( void )
//* Arguments
//*	char	data	command data(lower 4bits are abalable)
//*	char	mode	0:Commnad, 1:Data
//*	char	r_w	0:Wite,    1:Read
//* Return
//*	res	-1: error
//*		 0: normal end
//*
//******************************************************************************
int lcd_strobe( char data, char mode, char r_w )
{
	char	rs;
	char	rd;
	char	temp[3];
	char	data_int;
	uint8_t	res;
	char	str[80];

	rs = mode & 0x01;
	rd = (r_w & 0x01) << 1;

	
	data_int = ( (data & 0x0f) << 4 )  | rs | rd | Backlight;
	temp[0] = data_int;
	temp[1] = data_int | LCD_E;
	temp[2] = data_int;
	if( res=bcm2835_i2c_write( temp, 3 ) )
	{
		get_reaosn( res, str );
		printf( "Write error: %s[%d]\n", str, res );
		return( -1 );
	}
	return( 0 );
}

//******************************************************************************
//*	LCDW[ɃR}h݂̏s܂
//*
//* Function Name
//*	int lcd_write_command( char data )
//* Arguments
//*	char data	byte data
//* Return
//*	int res		0: normal end
//*			1: error
//*
//******************************************************************************
int lcd_write_command( char data )
{
	char	temp;
	int	ret;

	temp = (data >> 4) & 0x0f;
	ret  = lcd_strobe( temp,        LCD_COMMAND, LCD_WR );
	ret |= lcd_strobe( data & 0x0f, LCD_COMMAND, LCD_WR );
	bcm2835_delayMicroseconds( LCD_RECOVERY );
	return( ret );
}

//******************************************************************************
//*	LCDW[Ƀf[^݂̏s܂
//*
//* Function Name
//*	int lcd_write_data( char data )
//* Arguments
//*	char data	byte data
//* Return
//*	int res		0: normal end
//*			1: error
//*
//******************************************************************************
int lcd_write_data( char data )
{
	char	temp;
	int	ret;

	temp = (data >> 4) & 0x0f;
	ret  = lcd_strobe( temp,        LCD_DATA, LCD_WR );
	ret |= lcd_strobe( data & 0x0f, LCD_DATA, LCD_WR );
	bcm2835_delayMicroseconds( 40LL );
	return( ret );
}

//******************************************************************************
//*	LCDW[ɑ΂ĕ\ON/OFFs܂
//*
//* Function Name
//*	void lcd_on_off( int flag )
//* Arguments
//*	flag	0: OFF
//*		1: ON
//* Return
//*	none
//*
//******************************************************************************
void lcd_on_off( int flag )
{
	if( flag )	lcd_write_command( LCD_ON );
	else		lcd_write_command( LCD_OFF );
}

//******************************************************************************
//*	LCDW[ɑ΂ĕ\̃NAs܂
//*
//* Function Name
//*	void lcd_clear( void )
//* Arguments
//*	none
//* Return
//*	none
//*
//******************************************************************************
void lcd_clear( void )
{
	lcd_write_command( LCD_CLEAR );
	bcm2835_delay( 10 );
}

//******************************************************************************
//*	LCDW[ɑ΂ďs܂
//*
//* Function Name
//*	void lcd_init( void )
//* Arguments
//*	none
//* Return
//*	none
//*
//******************************************************************************
void lcd_init( void )
{

	// I/O s̏
	if( !bcm2835_i2c_begin() )
	{
		printf( "I2C initialize error!!\n" );
		exit(1);
	}

	// I2Cݒ
	bcm2835_i2c_setClockDivider( CLK_DIV );

	// tF[YP Data=0011
	lcd_strobe( 0x03, LCD_COMMAND, LCD_WR );
	bcm2835_delay( 5 );

	// tF[YQ Data=0011
	lcd_strobe( 0x03, LCD_COMMAND, LCD_WR );
	bcm2835_delayMicroseconds( 100LL );

	// tF[YR Data=0011
	lcd_strobe( 0x03, LCD_COMMAND, LCD_WR );
	bcm2835_delayMicroseconds( 100LL );

	// tF[YS Data=0010
	lcd_strobe( 0x02, LCD_COMMAND, LCD_WR );
	bcm2835_delayMicroseconds( 40LL );

	// t@NVݒ( 2 Line 5~10 dots)
	lcd_write_command( LCD_FUNCTION_SET );

	// Display clear
	// Display CLEAR & ON
	lcd_clear();
	lcd_on_off( ON );
}

//******************************************************************************
//*	I2CC^tF[XJ܂BB
//*
//* Function Name
//*	void lcd_close( void )
//* Arguments
//*	none
//* Return
//*	none
//*
//******************************************************************************
void lcd_close( void )
{
	bcm2835_i2c_end();
	bcm2835_close();
}

//******************************************************************************
//*	sԍLCDW[̍s擪̃AhXɕϊ
//*
//* Function Name
//*	int lcd_location( int line )
//* Arguments
//*	int line		target position 0:1st Line, 1:2nd Line, 2:3rd Line, 4:4th Line
//* Return
//*	ret:		0: normal end
//*			-1:error
//*
//******************************************************************************
int lcd_location( int line )
{
	char	addr;

	switch( line )
	{
	case 0:
		addr = 0x80;
		break;
	case 1:
		addr = 0x80 | 0x40;
		break;
	case 2:
		addr = 0x80 + 20;
		break;
	case 3:
		addr = (0x80 | 0x40) + 20;
		break;
	default:
		printf( "Illegal line number:[%d]\n", line );
		return( -1 );
	}
	
	lcd_write_command( addr );
	return( 0 );
}

//******************************************************************************
//*	LCDW[ɕ\܂B
//*
//* Function Name
//*	void lcd_print( int line, char *str )
//* Arguments
//*	int line		target position 0:first line, 1:second line
//*	char *str		string data
//* Return
//*	none
//*
//******************************************************************************
void lcd_print( int line, char *str )
{
	int	i;
	int	len;
	char	data;

	lcd_location( line );
	for( i=0; i<16; i++)
	{
		lcd_write_data( ' ' );
	}

	lcd_location( line );
	len = strlen(str);
	if( len > 16 ) len = 16;
	for( i=0; i<strlen(str); i++ )
	{
		lcd_write_data( str[i] );
	}
}

//******************************************************************************
//*	wv\
//*
//* Function Name
//*	void print_help( void )
//* Arguments
//*	mome
//* Return
//*	none
//*
//******************************************************************************
void print_help( void)
{
	 fprintf( stderr, "%s", 
		 "\nUsage: lcd_exp [[-a <addr>] [-i] [-h]  <message> [<line#>]"
		 "\n\n"
		 "********************** Parameters **********************\n"
		 "  message    : ASCII character string( include SJIS katakana)\n"
		 "  line#      : Line number to display 0 or 1(default:0)\n"

		 "Options:\n"
		 "  -a <addr>  : Target address on I2C bus.\n"
		 "  -i         : Initialize LCD\n"
		 "  -h         : Print this help\n"
		 "********************************************************\n"
		"\n" );
}

int main( int argc, char *argv[] )
{
	int	line = 0;
	int	c;
	int	init_flag = 0;

	I2C_addr = I2C_BASE_ADDR;

	// Get command options
	while( (c = getopt(argc, argv, "a:ih")) != -1 )
	{
		switch (c)
                {
		case 'a':
			I2C_addr = strtol( optarg, NULL, 0 );
			break;
		case 'i':
			init_flag++;
			break;
		case 'h':
			print_help();
			return(0);
		default:
			/*NOTREACHED*/
			;
		}
	}
	argc -= optind;
	argv += optind;

	if( argc == 2 ) line = (int)strtol( argv[1], NULL, 0 );

	// Cu
	if( !bcm2835_init() )
	{
		printf( "bcm2835 library initialize error!!\n" );
		return(1);
	}

	// I2Cݒ
	bcm2835_i2c_setSlaveAddress( I2C_addr );

	if( init_flag )
	{
		lcd_init();
	}

	lcd_print( line, argv[0] );
	bcm2835_close();
	return( 0 );

}
